@@ -1,15 +1,22 @@
-use super::errors::NovaError;
-use super::traits::{CompressedGroup, Group};
-use core::fmt::Debug;
-use core::ops::{Add, AddAssign, Mul, MulAssign};
+use super::{
+  errors::NovaError,
+  traits::{AbsorbInROTrait, AppendToTranscriptTrait, CompressedGroup, Group, HashFuncTrait},
+};
+use core::{
+  fmt::Debug,
+  marker::PhantomData,
+  ops::{Add, AddAssign, Mul, MulAssign},
+};
 use digest::{ExtendableOutput, Input};
+use ff::Field;
 use merlin::Transcript;
 use sha3::Shake256;
 use std::io::Read;
 
 #[derive(Debug)]
 pub struct CommitGens<G: Group> {
-  gens: Vec<G>,
+  gens: Vec<G::PreprocessedGroupElement>,
+  _p: PhantomData<G>,
 }
 
 #[derive(Clone, Copy, Debug, PartialEq, Eq)]
@@ -27,14 +34,17 @@ impl<G: Group> CommitGens<G> {
     let mut shake = Shake256::default();
     shake.input(label);
     let mut reader = shake.xof_result();
-    let mut gens: Vec<G> = Vec::new();
+    let mut gens: Vec<G::PreprocessedGroupElement> = Vec::new();
     let mut uniform_bytes = [0u8; 64];
     for _ in 0..n {
       reader.read_exact(&mut uniform_bytes).unwrap();
       gens.push(G::from_uniform_bytes(&uniform_bytes).unwrap());
     }
 
-    CommitGens { gens }
+    CommitGens {
+      gens,
+      _p: PhantomData::default(),
+    }
   }
 }
 
@@ -71,16 +81,25 @@ impl<G: Group> CommitTrait<G> for [G::Scalar] {
   }
 }
 
-pub trait AppendToTranscriptTrait {
-  fn append_to_transcript(&self, label: &'static [u8], transcript: &mut Transcript);
-}
-
 impl<G: Group> AppendToTranscriptTrait for Commitment<G> {
   fn append_to_transcript(&self, label: &'static [u8], transcript: &mut Transcript) {
     transcript.append_message(label, self.comm.compress().as_bytes());
   }
 }
 
+impl<G: Group> AbsorbInROTrait<G> for Commitment<G> {
+  fn absorb_in_ro(&self, ro: &mut G::HashFunc) {
+    let (x, y, is_infinity) = self.comm.to_coordinates();
+    ro.absorb(x);
+    ro.absorb(y);
+    ro.absorb(if is_infinity {
+      G::Base::one()
+    } else {
+      G::Base::zero()
+    });
+  }
+}
+
 impl<C: CompressedGroup> AppendToTranscriptTrait for CompressedCommitment<C> {
   fn append_to_transcript(&self, label: &'static [u8], transcript: &mut Transcript) {
     transcript.append_message(label, self.comm.as_bytes());
