@@ -3,26 +3,24 @@
 #![allow(clippy::type_complexity)]
 #![deny(missing_docs)]
 
-mod commitments;
-
 pub mod bellperson;
 mod circuit;
+mod commitments;
+mod constants;
 pub mod errors;
-mod gadgets;
+pub mod gadgets;
 pub mod pasta;
 mod poseidon;
 pub mod r1cs;
 pub mod traits;
 
-use std::marker::PhantomData;
-
-use commitments::{AppendToTranscriptTrait, CompressedCommitment};
+use commitments::CompressedCommitment;
 use errors::NovaError;
-use merlin::Transcript;
 use r1cs::{
   R1CSGens, R1CSInstance, R1CSShape, R1CSWitness, RelaxedR1CSInstance, RelaxedR1CSWitness,
 };
-use traits::{ChallengeTrait, Group};
+use std::marker::PhantomData;
+use traits::{AbsorbInROTrait, Group, HashFuncConstantsTrait, HashFuncTrait};
 
 /// A SNARK that holds the proof of a step of an incremental computation
 pub struct StepSNARK<G: Group> {
@@ -31,10 +29,6 @@ pub struct StepSNARK<G: Group> {
 }
 
 impl<G: Group> StepSNARK<G> {
-  fn protocol_name() -> &'static [u8] {
-    b"NovaStepSNARK"
-  }
-
   /// Takes as input a Relaxed R1CS instance-witness tuple `(U1, W1)` and
   /// an R1CS instance-witness tuple `(U2, W2)` with the same structure `shape`
   /// and defined with respect to the same `gens`, and outputs
@@ -48,7 +42,6 @@ impl<G: Group> StepSNARK<G> {
     W1: &RelaxedR1CSWitness<G>,
     U2: &R1CSInstance<G>,
     W2: &R1CSWitness<G>,
-    transcript: &mut Transcript,
   ) -> Result<
     (
       StepSNARK<G>,
@@ -56,18 +49,27 @@ impl<G: Group> StepSNARK<G> {
     ),
     NovaError,
   > {
-    // append the protocol name to the transcript
-    //transcript.append_protocol_name(StepSNARK::protocol_name());
-    transcript.append_message(b"protocol-name", StepSNARK::<G>::protocol_name());
+    // initialize a new RO
+    let mut ro = G::HashFunc::new(<<G as traits::Group>::HashFunc as HashFuncTrait<
+      G::Base,
+      G::Scalar,
+    >>::Constants::new());
+
+    // append S to the transcript
+    S.absorb_in_ro(&mut ro);
+
+    // append U1 and U2 to transcript
+    U1.absorb_in_ro(&mut ro);
+    U2.absorb_in_ro(&mut ro);
 
     // compute a commitment to the cross-term
     let (T, comm_T) = S.commit_T(gens, U1, W1, U2, W2)?;
 
     // append `comm_T` to the transcript and obtain a challenge
-    comm_T.append_to_transcript(b"comm_T", transcript);
+    comm_T.absorb_in_ro(&mut ro);
 
-    // compute a challenge from the transcript
-    let r = G::Scalar::challenge(b"r", transcript);
+    // compute a challenge from the RO
+    let r = ro.get_challenge();
 
     // fold the instance using `r` and `comm_T`
     let U = U1.fold(U2, &comm_T, &r)?;
@@ -78,7 +80,7 @@ impl<G: Group> StepSNARK<G> {
     // return the folded instance and witness
     Ok((
       StepSNARK {
-        comm_T,
+        comm_T: comm_T.compress(),
         _p: Default::default(),
       },
       (U, W),
@@ -92,21 +94,32 @@ impl<G: Group> StepSNARK<G> {
   /// if and only if `U1` and `U2` are satisfiable.
   pub fn verify(
     &self,
+    S: &R1CSShape<G>,
     U1: &RelaxedR1CSInstance<G>,
     U2: &R1CSInstance<G>,
-    transcript: &mut Transcript,
   ) -> Result<RelaxedR1CSInstance<G>, NovaError> {
-    // append the protocol name to the transcript
-    transcript.append_message(b"protocol-name", StepSNARK::<G>::protocol_name());
+    // initialize a new RO
+    let mut ro = G::HashFunc::new(<<G as traits::Group>::HashFunc as HashFuncTrait<
+      G::Base,
+      G::Scalar,
+    >>::Constants::new());
+
+    // append S to the transcript
+    S.absorb_in_ro(&mut ro);
+
+    // append U1 and U2 to transcript
+    U1.absorb_in_ro(&mut ro);
+    U2.absorb_in_ro(&mut ro);
 
     // append `comm_T` to the transcript and obtain a challenge
-    self.comm_T.append_to_transcript(b"comm_T", transcript);
+    let comm_T = self.comm_T.decompress()?;
+    comm_T.absorb_in_ro(&mut ro);
 
-    // compute a challenge from the transcript
-    let r = G::Scalar::challenge(b"r", transcript);
+    // compute a challenge from the RO
+    let r = ro.get_challenge();
 
     // fold the instance using `r` and `comm_T`
-    let U = U1.fold(U2, &self.comm_T, &r)?;
+    let U = U1.fold(U2, &comm_T, &r)?;
 
     // return the folded instance
     Ok(U)
@@ -139,12 +152,133 @@ impl<G: Group> FinalSNARK<G> {
 #[cfg(test)]
 mod tests {
   use super::*;
-  use crate::traits::PrimeField;
+  use ::bellperson::{gadgets::num::AllocatedNum, ConstraintSystem, SynthesisError};
+  use ff::{Field, PrimeField};
   use rand::rngs::OsRng;
 
   type S = pasta_curves::pallas::Scalar;
   type G = pasta_curves::pallas::Point;
 
+  fn synthesize_tiny_r1cs_bellperson<Scalar: PrimeField, CS: ConstraintSystem<Scalar>>(
+    cs: &mut CS,
+    x_val: Option<Scalar>,
+  ) -> Result<(), SynthesisError> {
+    // Consider a cubic equation: `x^3 + x + 5 = y`, where `x` and `y` are respectively the input and output.
+    let x = AllocatedNum::alloc(cs.namespace(|| "x"), || Ok(x_val.unwrap()))?;
+    let _ = x.inputize(cs.namespace(|| "x is input"));
+
+    let x_sq = x.square(cs.namespace(|| "x_sq"))?;
+    let x_cu = x_sq.mul(cs.namespace(|| "x_cu"), &x)?;
+    let y = AllocatedNum::alloc(cs.namespace(|| "y"), || {
+      Ok(x_cu.get_value().unwrap() + x.get_value().unwrap() + Scalar::from(5u64))
+    })?;
+    let _ = y.inputize(cs.namespace(|| "y is output"));
+
+    cs.enforce(
+      || "y = x^3 + x + 5",
+      |lc| {
+        lc + x_cu.get_variable()
+          + x.get_variable()
+          + CS::one()
+          + CS::one()
+          + CS::one()
+          + CS::one()
+          + CS::one()
+      },
+      |lc| lc + CS::one(),
+      |lc| lc + y.get_variable(),
+    );
+
+    Ok(())
+  }
+
+  #[test]
+  fn test_tiny_r1cs_bellperson() {
+    use super::bellperson::{
+      r1cs::{NovaShape, NovaWitness},
+      shape_cs::ShapeCS,
+      solver::SatisfyingAssignment,
+    };
+
+    // First create the shape
+    let mut cs: ShapeCS<G> = ShapeCS::new();
+    let _ = synthesize_tiny_r1cs_bellperson(&mut cs, None);
+    let shape = cs.r1cs_shape();
+    let gens = cs.r1cs_gens();
+
+    // Now get the instance and assignment for one instance
+    let mut cs: SatisfyingAssignment<G> = SatisfyingAssignment::new();
+    let _ = synthesize_tiny_r1cs_bellperson(&mut cs, Some(S::from(5)));
+    let (U1, W1) = cs.r1cs_instance_and_witness(&shape, &gens).unwrap();
+
+    // Make sure that the first instance is satisfiable
+    assert!(shape.is_sat(&gens, &U1, &W1).is_ok());
+
+    // Now get the instance and assignment for second instance
+    let mut cs: SatisfyingAssignment<G> = SatisfyingAssignment::new();
+    let _ = synthesize_tiny_r1cs_bellperson(&mut cs, Some(S::from(135)));
+    let (U2, W2) = cs.r1cs_instance_and_witness(&shape, &gens).unwrap();
+
+    // Make sure that the second instance is satisfiable
+    assert!(shape.is_sat(&gens, &U2, &W2).is_ok());
+
+    // execute a sequence of folds
+    execute_sequence(&gens, &shape, &U1, &W1, &U2, &W2);
+  }
+
+  fn execute_sequence(
+    gens: &R1CSGens<G>,
+    shape: &R1CSShape<G>,
+    U1: &R1CSInstance<G>,
+    W1: &R1CSWitness<G>,
+    U2: &R1CSInstance<G>,
+    W2: &R1CSWitness<G>,
+  ) {
+    // produce a default running instance
+    let mut r_W = RelaxedR1CSWitness::default(shape);
+    let mut r_U = RelaxedR1CSInstance::default(gens, shape);
+
+    // produce a step SNARK with (W1, U1) as the first incoming witness-instance pair
+    let res = StepSNARK::prove(gens, shape, &r_U, &r_W, U1, W1);
+    assert!(res.is_ok());
+    let (step_snark, (_U, W)) = res.unwrap();
+
+    // verify the step SNARK with U1 as the first incoming instance
+    let res = step_snark.verify(shape, &r_U, U1);
+    assert!(res.is_ok());
+    let U = res.unwrap();
+
+    assert_eq!(U, _U);
+
+    // update the running witness and instance
+    r_W = W;
+    r_U = U;
+
+    // produce a step SNARK with (W2, U2) as the second incoming witness-instance pair
+    let res = StepSNARK::prove(gens, shape, &r_U, &r_W, U2, W2);
+    assert!(res.is_ok());
+    let (step_snark, (_U, W)) = res.unwrap();
+
+    // verify the step SNARK with U1 as the first incoming instance
+    let res = step_snark.verify(shape, &r_U, U2);
+    assert!(res.is_ok());
+    let U = res.unwrap();
+
+    assert_eq!(U, _U);
+
+    // update the running witness and instance
+    r_W = W;
+    r_U = U;
+
+    // produce a final SNARK
+    let res = FinalSNARK::prove(&r_W);
+    assert!(res.is_ok());
+    let final_snark = res.unwrap();
+    // verify the final SNARK
+    let res = final_snark.verify(gens, shape, &r_U);
+    assert!(res.is_ok());
+  }
+
   #[test]
   fn test_tiny_r1cs() {
     let one = S::one();
@@ -247,50 +381,7 @@ mod tests {
     let (O, U1, W1) = rand_inst_witness_generator(&gens, &I);
     let (_O, U2, W2) = rand_inst_witness_generator(&gens, &O);
 
-    // produce a default running instance
-    let mut r_W = RelaxedR1CSWitness::default(&S);
-    let mut r_U = RelaxedR1CSInstance::default(&gens, &S);
-
-    // produce a step SNARK with (W1, U1) as the first incoming witness-instance pair
-    let mut prover_transcript = Transcript::new(b"StepSNARKExample");
-    let res = StepSNARK::prove(&gens, &S, &r_U, &r_W, &U1, &W1, &mut prover_transcript);
-    assert!(res.is_ok());
-    let (step_snark, (_U, W)) = res.unwrap();
-
-    // verify the step SNARK with U1 as the first incoming instance
-    let mut verifier_transcript = Transcript::new(b"StepSNARKExample");
-    let res = step_snark.verify(&r_U, &U1, &mut verifier_transcript);
-    assert!(res.is_ok());
-    let U = res.unwrap();
-
-    assert_eq!(U, _U);
-
-    // update the running witness and instance
-    r_W = W;
-    r_U = U;
-
-    // produce a step SNARK with (W2, U2) as the second incoming witness-instance pair
-    let res = StepSNARK::prove(&gens, &S, &r_U, &r_W, &U2, &W2, &mut prover_transcript);
-    assert!(res.is_ok());
-    let (step_snark, (_U, W)) = res.unwrap();
-
-    // verify the step SNARK with U1 as the first incoming instance
-    let res = step_snark.verify(&r_U, &U2, &mut verifier_transcript);
-    assert!(res.is_ok());
-    let U = res.unwrap();
-
-    assert_eq!(U, _U);
-
-    // update the running witness and instance
-    r_W = W;
-    r_U = U;
-
-    // produce a final SNARK
-    let res = FinalSNARK::prove(&r_W);
-    assert!(res.is_ok());
-    let final_snark = res.unwrap();
-    // verify the final SNARK
-    let res = final_snark.verify(&gens, &S, &r_U);
-    assert!(res.is_ok());
+    // execute a sequence of folds
+    execute_sequence(&gens, &S, &U1, &W1, &U2, &W2);
   }
 }
